/***************************** C HEADER FILE *********************************
 **
 **  Project:    Platform SDK - Module header
 **  Filename:   MODULE.H
 **  Version:    1.00
 **  Date:       September 22, 2017
 **
 ******************************************************************************
 **
 *****************************************************************************/
#ifndef _MODULE__H_
#define _MODULE__H_


/*****************************************************************************
 **
 **    MODULES USED
 **
 *****************************************************************************/
#include <time.h>

/////////////////////////////////////////////////////////////////////////////
///
/// @defgroup group_def Definitions and Macros
/// @{
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
///
/// @name Declarations
/// @{
/////////////////////////////////////////////////////////////////////////////

/// Define to include in the declaration of functions, in case the calling conventions change.
#define MODDECL
/// @}

/////////////////////////////////////////////////////////////////////////////
///
/// @name Data types
/// These defines provide platform independent data types.
/// @{
/////////////////////////////////////////////////////////////////////////////

/// Signed 8 bit integer, corresponding to SINT in VPL.
#define int8                           signed char
/// Signed 16 bit integer, corresponding to INT in VPL.
#define int16                          signed short
/// Signed 32 bit integer, corresponding to DINT in VPL.
#define int32                          signed long
/// Unsigned 8 bit integer.
#define uint8                          unsigned char
/// Unsigned 16 bit integer.
#define uint16                         unsigned short
/// Unsigned 32 bit integer.
#define uint32                         unsigned long
/// Handle to internal system data.
#define HANDLE                         void
/// @}

/////////////////////////////////////////////////////////////////////////////
///
/// @name  SOS Flags
/// Flags for the SOS objects. These flags can be combined bitwise.
/// @{
/////////////////////////////////////////////////////////////////////////////

/// No special flags
#define SOS_FLAG_NONE                  0x00
/// Read-only object. The object can not be written by external systems, such as the RTCU IDE.
#define SOS_FLAG_RO                    0x01
/// The object is encrypted. Number and boolean objects can not be encrypted.
#define SOS_FLAG_CRYPT                 0x02
///@}

/////////////////////////////////////////////////////////////////////////////
///
/// @name Notification events
/// Event IDs for notifications sent to moduleNotify
/// @{
/////////////////////////////////////////////////////////////////////////////

/// The system is halting
#define EVENT_HALT                     1
/// The system is resetting
#define EVENT_RESET                    2
/// The system is shutting down
#define EVENT_SHUTDOWN                 3
/// External power has disappeared/has appeared again. Parm is 1 when the power disappears and 0 when it appears again.
#define EVENT_POWERFAIL                4
/// The system is entering/leaving pmWaitEvent. Parm is 1 when entering pmWaitEvent and 0 when leaving.
#define EVENT_POWERSAVE                5
/// @} Notification events

/////////////////////////////////////////////////////////////////////////////
///
/// @name Alignment macros
/// Macros for working with unaligned data
/// @{
/////////////////////////////////////////////////////////////////////////////

/// Attribute to set alignment
#define ALIGN_ATTR                     __attribute__ ((packed, aligned(2)))

#define PACKED_ATTR                    __attribute__ ((packed))

/// Macro for unaligned read of a 16 bit value
/// @param addr The address to read from
/// @return The value read from the address
#define read16b(addr)                  ((uint16)(((((uint8*)(addr))[1]<<8) | ((uint8*)(addr))[0])))
/// Macro for unaligned read of a 24 bit value
/// @param addr The address to read from
/// @return The value read from the address
#define read24b(addr)                  ((uint32)((((uint8*)(addr))[2]<<16) | (((uint8*)(addr))[1]<<8)  | (((uint8*)(addr))[0])))
/// Macro for unaligned read of a 32 bit value
/// @param addr The address to read from
/// @return The value read from the address
#define read32b(addr)                  ((uint32)((((uint8*)(addr))[3]<<24) | (((uint8*)(addr))[2]<<16) | (((uint8*)(addr))[1]<<8) | (((uint8*)(addr))[0])))
/// Macro for unaligned write of a 16 bit value
/// @param addr The address to write to
/// @param val The value to write
#define write16b(addr, val)            { ((uint8*)(addr))[0] = ((val)&0xff); ((uint8*)(addr))[1] = ((val)>>8)&0xff; }
/// Macro for unaligned write of a 24 bit value
/// @param addr The address to write to
/// @param val The value to write
#define write24b(addr, val)            { ((uint8*)(addr))[0] = ((val)&0xff); ((uint8*)(addr))[1] = ((val)>>8)&0xff; ((uint8*)(addr))[2] = ((val)>>16)&0xff; }
/// Macro for unaligned write of a 32 bit value
/// @param addr The address to write to
/// @param val The value to write
#define write32b(addr, val)            { ((uint8*)(addr))[0] = ((val)&0xff); ((uint8*)(addr))[1] = ((val)>>8)&0xff; ((uint8*)(addr))[2] = ((val)>>16)&0xff; ((uint8*)(addr))[3] = ((val)>>24)&0xff; }

/// @} Alignment macros

/////////////////////////////////////////////////////////////////////////////
///
/// @name Enumerations
/// Enumeration data types.
/// @{
/////////////////////////////////////////////////////////////////////////////



/////////////////////////////////////////////////////////////////////////////
///
/// IO Signals that can be controlled using vplSetIOSignal()
///
/// Note that not all signals are available on all targets.
///
/////////////////////////////////////////////////////////////////////////////
typedef enum {
   /// Control power to the optional RF module. Set to 0 to turn the power on.
   RF_OFF = 1,
   /// Control the CFG pin on the RF module.
   RF_CFG,
   /// Enable serial port 0 if available
   SER0_EN,
   /// Enable serial port 1 if available
   SER1_EN,
   /// Enable serial port 2 if available
   SER2_EN,
   /// Enable serial port 3 if available
   SER3_EN,
   /// Enable serial port 4 if available (reserved)
   SER4_EN,
   /// Enable serial port 5 if available (reserved)
   SER5_EN,
   /// Enable serial port 6 if available (reserved)
   SER6_EN,
   /// Enable serial port 7 if available (reserved)
   SER7_EN,
   /// Enable serial port 8 if available (reserved)
   SER8_EN,
} vpl_io_signals;

/// @} Enumerations

/// @} group_def

/////////////////////////////////////////////////////////////////////////////
///
/// @defgroup group_typedef   Type definitions
/// @{
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
/// Function footprint definition
///
/// @param [in] pCPU               The system handle.
/// @param [in] pBase              Pointer to structure with function parameters.
///
///
/// @return
///   Return value to pass on to VPL.
///
/////////////////////////////////////////////////////////////////////////////
typedef int32 (MODDECL *vpl_function_call)(HANDLE* pCPU, void* pbase);



/////////////////////////////////////////////////////////////////////////////
/// Audio capture callback, for use with vplAudioStreamRegister().
/// When this callback is called, the module must fill the buffer with the data to play and store the data size in size.
///
/// @param [in]  id                 The ID of the stream to provide data for.
/// @param [in]  buffer             The buffer to place the data in.
/// @param [in]  buffer_size        The maximum size of data to write to the buffer.
/// @param [out] size               The actual amount of data written to the buffer.
///
/// @retval   0                 Success.
/// @retval  <0                 Error, currently ignored.
///
/////////////////////////////////////////////////////////////////////////////
typedef int (*capture_handle)(int16 id, char* buffer, uint32 buffer_size, uint32* size);

/////////////////////////////////////////////////////////////////////////////
/// Audio playback callback, for use with vplAudioStreamRegister().
/// When this callback is called, the module must use the data provided in the buffer and return.
///
/// @param [in]  id                 The ID of the stream to provide data for.
/// @param [in]  buffer             The buffer containing the audio data.
/// @param [in]  size               The amount of data in the buffer.
///
/// @retval   0                 Success.
/// @retval  <0                 Error, currently ignored.
///
/////////////////////////////////////////////////////////////////////////////
typedef int (*playback_handle)(int16 id, char* data, uint32 size);


/// @} group_typedef

/////////////////////////////////////////////////////////////////////////////
/// Function table entry
///
/// Parameters:
/// @param  fname              The name of the function.
///                        Used in VPL to identify the function
/// @param  func               Pointer to the function in module.
///
/////////////////////////////////////////////////////////////////////////////
typedef struct
{
   const char* fname; // function name
   vpl_function_call func; // function address.
} vplFunctionEntry;

/****************************************************************************
 **
 **    EXPORTED VARIABLES
 **
 ****************************************************************************/

/////////////////////////////////////////////////////////////////////////////
///
/// @defgroup group_func Functions
/// The following functions can be called from the Extension Module
///  to interact with the system.
/// @{
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
/// @name Basic functions
/// @{
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
/// This function must be called from moduleInit to register the functions in the system,
/// making it possible to call the functions from VPL.
/// It is possible to call this function multiple times, to install functions
/// under multiple module names or to dynamically choose which functions to install.
///
/// @param [in]  modname            The name of the module to install the function for.
///                                 The first character must be a letter or underscore, the remaining characters may also contain digits.
///                                 Maximum length is 20 characters.
/// @param [in]  ftable             Pointer to the Function table of the module.
///                                   The table must be terminated with an empty entry,
///                                   where both fname and func are NULL.
///
/// @retval   0                 Success.
/// @retval   1                 A function with the same name already exists.
/// @retval   2                 Invalid module name.
/// @retval   3                 Modules are not supported.
///
/////////////////////////////////////////////////////////////////////////////
uint16 MODDECL vplInstallFunctions(
                                   char modname[16],
                                   const vplFunctionEntry* ftable
                                   );

/////////////////////////////////////////////////////////////////////////////
/// Send a debug message to Device output in a connected IDE.
/// The text can be 480 characters long.
/// This function can be used for debugging the module,
/// by printing out strategic locations and variables.
///
///@param [in]   message            The text to send.
///
///
/////////////////////////////////////////////////////////////////////////////
void MODDECL vplDebug(const char* message);

/// @} common functions

/////////////////////////////////////////////////////////////////////////////
/// @name String functions
/// The following functions are used to move string variables between VPL and C.
/// @{
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
/// Reads a string from VPL and returns a pointer to it.
/// A pointer to an empty string is returned if the string handle is not valid.
///
/// @param[in] cpu                The handle to the system.
/// @param[in] stringid           The handle to the string.
///
///
/// @return
///   A pointer to the string.
///
/////////////////////////////////////////////////////////////////////////////
const char* MODDECL vplStringGet(
                                 HANDLE* cpu,
                                 uint16 stringid
                                 );

/////////////////////////////////////////////////////////////////////////////
/// Create a new VPL string from the provided C string.
///
/// @param[in]   cpu                The handle to the system.
/// @param[in]   text               The text to store in the new VPL string.
///
///
/// @return
///   Handle to the new string.
///
/////////////////////////////////////////////////////////////////////////////
uint16 MODDECL vplStringMake(HANDLE* cpu, const char* text);

/////////////////////////////////////////////////////////////////////////////
/// Used to update existing VPL strings by releasing the old string and
/// returning a handle to the new string that should be used instead.
/// Mainly for use with ACCESS variables.
///
///  @param[in]  cpu                The handle to the system.
///  @param[in]  strid              The handle to the old VPL string. It is invalid after this call.
///  @param[in]  text               The text to store in the new VPL string.
///
///
/// @return
///   Handle to the new VPL string.
///
/////////////////////////////////////////////////////////////////////////////
uint16 MODDECL vplStringUpdate(HANDLE* cpu, uint16 strid, const char* text);

/// @}

/////////////////////////////////////////////////////////////////////////////
///
/// @name Time functions
/// The following functions are used to work with the time.
/// @{
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
/// Retrieves the current system time
///
///
/// @return
///   The current time as linsec.
///
/////////////////////////////////////////////////////////////////////////////
int32 MODDECL vplClockGet(void);

/////////////////////////////////////////////////////////////////////////////
/// Converts a time provided as the individual elements to a linsec
///
/// @param[in]  year               The year.   (1980..2048)
/// @param[in]  month              The month.  (1..12)
/// @param[in]  day                The day.    (1..31)
/// @param[in]  hour               The hour.   (0..23)
/// @param[in]  minute             The minute. (0..59)
/// @param[in]  second             The second. (0..59)
///
///
/// @return
///   The time as linsec.
///
/////////////////////////////////////////////////////////////////////////////
int32 MODDECL vplClockTimeToLinsec(
                                   int16 year,
                                   int8 month,
                                   int8 day,
                                   int8 hour,
                                   int8 minute,
                                   int8 second
                                   );

/////////////////////////////////////////////////////////////////////////////
/// Convert linsec to the individual time elements.
///
///   @param[in] linsec             The linsec time
///
/// @param[out]  year               The year.
/// @param[out]  month              The month.
/// @param[out]  day                The day.
/// @param[out]  hour               The hour.
/// @param[out]  minute             The minute.
/// @param[out]  second             The second.
///
///
/////////////////////////////////////////////////////////////////////////////
void MODDECL vplClockLinsecToTime(
                                  int32 linsec,
                                  int16* year,
                                  int8* month,
                                  int8* day,
                                  int8* hour,
                                  int8* minute,
                                  int8* second
                                  );

/////////////////////////////////////////////////////////////////////////////
/// Convert a struct tm to linsec
///
/// @param[in]  time               Pointer to a struct tm with the time.
///
///
/// @return
///   The time as linsec.
///
/////////////////////////////////////////////////////////////////////////////
int32 MODDECL vplClockTMToLinsec(struct tm * time);

/////////////////////////////////////////////////////////////////////////////
/// Convert linsec to a struct tm
///
/// @param[in]  linsec             The linsec time
///
/// @param[out] time               Pointer to a struct tm with the time.
///
///
/////////////////////////////////////////////////////////////////////////////
void MODDECL vplClockLinsecToTM(int32 linsec, struct tm * time);

/// @}

/////////////////////////////////////////////////////////////////////////////
//
/// @name SOS functions
/// The following functions are used to work with the System Object Storage,
/// which can be used for storing settings for the module.
/// The functions provide the same functionality as the VPL functions, but with the added support
/// for using encrypted objects and objects that can not be changed from the IDE.
///@{
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
/// Create a data object in the SOS table
///
/// @param [in]  name               The name of the object.
/// @param [in]  data               The data of the object.
/// @param [in]  len                The number of bytes in the data.
/// @param [in]  flags              The flags for the object.
/// @param [in]  desc               A short description of the object. (max 80 characters + null terminator)
/// @param [in]  limit              The maximum number of bytes allowed in the data.
///
///
///  @retval 0                  Success.
///  @retval -1                 Error.
///  @retval -2                 Illegal parameter.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosDataCreate(
                             const char* name,
                             char* data,
                             int len,
                             int flags,
                             char* desc,
                             int limit
                             );

/////////////////////////////////////////////////////////////////////////////
/// Update a data object in the SOS table
///
/// @param [in]  name               The name of the object.
/// @param [in]  data               The data of the object.
/// @param [in]  len                The number of bytes in the data.
///
///
/// @retval   0              Success.
/// @retval  -1              Error.
/// @retval  -2              Illegal parameter.
/// @retval -10              Object is not found.
/// @retval -11              Object is not a data value.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosDataUpdate(const char* name, char* data, int len);

/////////////////////////////////////////////////////////////////////////////
/// Read a data object in the SOS table
///
///
/// @param [in]  name               The name of the object.
/// @param [in]  max_len            The maximum number of bytes that can be read.
///
/// @param [out]  data              The data of the object.
/// @param [out]  len               The number of bytes in the data.
/// @param [out]  flags             The flags for the object.
/// @param [out]  desc              A short description of the object. (max 80 characters + null terminator)
/// @param [out]  limit             The maximum number of bytes allowed in the data.
///
/// @retval  0                 Success.
/// @retval -1                 Error.
/// @retval -2                 Illegal parameter.
/// @retval -10                Object is not found.
/// @retval -11                Object is not a data value.
/// @retval -12                Data buffer is too small.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosDataGet(
                          const char* name,
                          char* data,
                          int max_len,
                          int* len,
                          int* flags,
                          char* desc,
                          int* limit
                          );

/////////////////////////////////////////////////////////////////////////////
/// Create a string object in the SOS table
///
/// @param [in]  name               The name of the object.
/// @param [in]  data               The string data of the object.
/// @param [in]  flags              The flags for the object.
/// @param [in]  desc               A short description of the object. (max 80 characters + null terminator)
/// @param [in]  limit              The maximum string length allowed.
///
///
/// @retval  0                 Success.
/// @retval -1                 Error.
/// @retval -2                 Illegal parameter.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosStringCreate(
                               const char* name,
                               const char* data,
                               int flags,
                               char* desc,
                               int limit
                               );

/////////////////////////////////////////////////////////////////////////////
/// Update a string object in the SOS table
///
/// @param [in]  name               The name of the object.
/// @param [in]  data               The string data of the object.
///
/// @retval   0                 Success.
/// @retval  -1                 Error.
/// @retval  -2                 Illegal parameter.
/// @retval  -10                Object is not found.
/// @retval  -11                Object is not a string value.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosStringUpdate(const char* name, const char* data);

/////////////////////////////////////////////////////////////////////////////
/// Read a string object in the SOS table
///
/// @param [in]  name               The name of the object.
/// @param [in]  max_len            The maximum number of bytes that can be read.
///
/// @param [out]  data              The data of the object.
/// @param [out]  flags             The flags for the object.
/// @param [out]  desc              A short description of the object. (max 80 characters + null terminator)
/// @param [out]  limit             The maximum number of bytes allowed in the string.
///
/// @retval  0                  Success.
/// @retval  -1                 Error.
/// @retval  -2                 Illegal parameter.
/// @retval  -10                Object is not found.
/// @retval  -11                Object is not a string value.
/// @retval  -12                Data buffer is too small.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosStringGet(
                            const char* name,
                            char* data,
                            int max_len,
                            int* flags,
                            char* desc,
                            int* limit
                            );

/////////////////////////////////////////////////////////////////////////////
/// Create a float object in the SOS table
///
/// @param [in]  name               The name of the object.
/// @param [in]  data               The float data of the object.
/// @param [in]  flags              The flags for the object.
/// @param [in]  desc               A short description of the object. (max 80 characters + null terminator)
/// @param [in]  min_data           The manimum value allowed.
/// @param [in]  max_data           The maximum value allowed.
///
///
/// @retval   0                 Success.
/// @retval  -1                 Error.
/// @retval  -2                 Illegal parameter.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosFloatCreate(
                              const char* name,
                              float data,
                              int flags,
                              char* desc,
                              float min_data,
                              float max_data
                              );

/////////////////////////////////////////////////////////////////////////////
/// Update a float object in the SOS table
///
/// @param [in]  name               The name of the object.
/// @param [in]  data               The value to store.
///
///
/// @retval    0                Success.
/// @retval   -1                Error.
/// @retval   -2                Illegal parameter.
/// @retval  -10                Object is not found.
/// @retval  -11                Object is not a float value.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosFloatUpdate(const char* name, float data);

/////////////////////////////////////////////////////////////////////////////
/// Read a float object in the SOS table
///
/// @param [in]  name               The name of the object.
///
/// @param [out]  data              The value of the object.
/// @param [out]  flags             The flags for the object.
/// @param [out]  desc              A short description of the object. (max 80 characters + null terminator)
/// @param [out]  min_data          The manimum value allowed.
/// @param [out]  max_data          The maximum value allowed.
///
/// @retval  0                  Success.
/// @retval  -1                 Error.
/// @retval  -2                 Illegal parameter.
/// @retval  -10                Object is not found.
/// @retval  -11                Object is not a float value.
/// @retval  -12                Data buffer is too small.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosFloatGet(
                           const char* name,
                           float* data,
                           int* flags,
                           char* desc,
                           float* min_data,
                           float* max_data
                           );

/////////////////////////////////////////////////////////////////////////////
/// Create an integer object in the SOS table
///
/// @param [in]  name           The name of the object.
/// @param [in]  data           The value to store.
/// @param [in]  flags          The flags for the object.
/// @param [in]  desc           A short description of the object. (max 80 characters + null terminator)
/// @param [in]  min_data       The manimum value allowed.
/// @param [in]  max_data       The maximum value allowed.
///
///
/// @retval  0                 Success.
/// @retval -1                 Error.
/// @retval -2                 Illegal parameter.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosIntCreate(
                            const char* name,
                            int data,
                            int flags,
                            char* desc,
                            int min_data,
                            int max_data
                            );

/////////////////////////////////////////////////////////////////////////////
/// Update an integer object in the SOS table
///
/// @param [in]  name           The name of the object.
/// @param [in]  data           The value to store.
///
/// @retval  0                  Success.
/// @retval  -1                 Error.
/// @retval  -2                 Illegal parameter.
/// @retval  -10                Object is not found.
/// @retval  -11                Object is not a integer value.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosIntUpdate(const char* name, int data);

/////////////////////////////////////////////////////////////////////////////
/// Read an integer object in the SOS table
///
/// @param [in]  name               The name of the object.
///
/// @param [out]  data              The value of the object.
/// @param [out]  flags             The flags for the object.
/// @param [out]  desc              A short description of the object. (max 80 characters + null terminator)
/// @param [out]  min_data          The manimum value allowed.
/// @param [out]  max_data          The maximum value allowed.
///
/// @retval   0                 Success.
/// @retval  -1                 Error.
/// @retval  -2                 Illegal parameter.
/// @retval  -10                Object is not found.
/// @retval  -11                Object is not a integer value.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosIntGet(
                         const char* name,
                         int* data,
                         int* flags,
                         char* desc,
                         int* min_data,
                         int* max_data
                         );

/////////////////////////////////////////////////////////////////////////////
/// Create a boolean object in the SOS table
///
/// @param [in]  name               The name of the object.
/// @param [in]  data               The value to store.
/// @param [in]  flags              The flags for the object.
/// @param [in]  desc               A short description of the object. (max 80 characters + null terminator)
///
///
/// @retval   0                 Success.
/// @retval  -1                 Error.
/// @retval  -2                 Illegal parameter.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosBoolCreate(const char* name, int data, int flags, char* desc);

/////////////////////////////////////////////////////////////////////////////
/// Update a boolean object in the SOS table
///
/// @param [in]  name           The name of the object.
/// @param [in]  data           The value to store.
///
///
///
/// @retval   0                 Success.
/// @retval  -1                 Error.
/// @retval  -2                 Illegal parameter.
/// @retval  -10                Object is not found.
/// @retval  -11                Object is not a boolean value.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosBoolUpdate(const char* name, int data);

/////////////////////////////////////////////////////////////////////////////
/// Read a boolean object in the SOS table
///
/// @param [in]  name               The name of the object.
///
/// @param [out]  data               The value of the object.
/// @param [out]  flags              The flags for the object.
/// @param [out]  desc               A short description of the object. (max 80 characters + null terminator)
///
/// @retval   0                 Success.
/// @retval  -1                 Error.
/// @retval  -2                 Illegal parameter.
/// @retval  -10                Object is not found.
/// @retval  -11                Object is not a boolean value.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosBoolGet(const char* name, int* data, int* flags, char* desc);

/////////////////////////////////////////////////////////////////////////////
/// Delete an object in the SOS table
///
/// @param [in]  name               The name of the object.
///
/// @retval   0                 Success.
/// @retval  -1                 Error.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSosDelete(const char* name);

/// @}

/////////////////////////////////////////////////////////////////////////////
//
/// @name Audio functions
/// The following functions are used to work with audio.<br/>
/// The audio format is in stereo, 16bit signed, little endian, at 16kHz.<br/>
/// It is stored interleaved, with first the left channel followed by the right channel, i.e. the byte stream looks like this:
/// "LLRRLLRRLLRR"
///@{
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
/// Register an audio stream handler.
/// Once a stream handler has been registered, it can be accessed from VPL using the audioRoute function.
///
/// @param [in]  id                 The ID of the stream to modify.
/// @param [in]  capt               The capture callback. It is called to read the next audio data.
/// @param [in]  play               The playback callback. It is called with the latest audio data, which must be handled.
///
/// @retval   0                 Success.
/// @retval  -1                 Invalid ID.
/// @retval  -2                 The stream is in use.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplAudioStreamRegister(int16 id, capture_handle capt, playback_handle play);


/// @}

/////////////////////////////////////////////////////////////////////////////
//
/// @name File functions
/// The following functions are used to work with files.
///@{
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
/// Map a VPL or device file name to the real file name.
/// Absolute and relative paths are supported.
/// Device paths start with "dev:".<br/>
/// Currently supported devices:<br/>
/// Serial ports: "dev:ser<n>" where n = 0..3
///
/// @param [in]  dest               The buffer to store the file name in.
/// @param [in]  dest_size          The size of the buffer.
/// @param [in]  file_format        The name of the file to convert. Supports using sprintf formatting such as "%d".
/// @param [in]  ...                Formatting arguments for file_format.
///
///   @retval   0 Success.
///   @retval  -1 Invalid format.
///   @retval  -2 Dest too small.
///   @retval  -3 Unknown device
///   @retval  -4 Invalid filename
///   @retval  -5 Invalid path
///   @retval  -6 Drive not available
///   @retval  -7 Busy
///   @retval  -10 Out of memory
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplFsMapFile(char* dest, size_t dest_size, const char* file_format, ...);


/// @}

/////////////////////////////////////////////////////////////////////////////
//
/// @name IO functions
/// The following functions are used to work with input and output signals.
///@{
/////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////////
/// Use this function to set the value of raw IO signals.
/// See the #vpl_io_signals enumeration for a list of the available signals.<br/>
/// <br/>
/// Note that not all signals are available on all targets.<br/>
/// Some signals, such as the serial port enable signals, may return success even when they are not present on the specific target.<br/>
/// This is for consistency with other targets that have such a signal.<br/>
///
/// @param [in]  io                 The ID of the signal to set
/// @param [in]  val                The value to set the pin to. Set it to 1 to assert the signal and to 0 to deassert it.
///
///
///   @retval   0 Success.
///   @retval  -1 Invalid signal.
///
/////////////////////////////////////////////////////////////////////////////
int MODDECL vplSetIOSignal(vpl_io_signals io, int val);


/// @}
/// @} group_func



#endif  /* _MODULE__H_ */
/****************************************************************************
 **
 **    EOF
 **
 ****************************************************************************/
