/***************************** C SOURCE FILE **********************************
 **
 **  Project:    VPL extension module
 **  Filename:   helloworld.c
 **
 ******************************************************************************
 **
 **  This exports a function which returns a string with 'Hello World'
 **
 *****************************************************************************/

/******************************************************************************
 **
 ** MODULES USED
 **
 *****************************************************************************/
#include <module.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>


/******************************************************************************
 **
 **    DEFINITIONS AND MACROS
 **
 *****************************************************************************/
#define WAV_HEADER_LEN 44

/******************************************************************************
 **
 **    TYPEDEFS AND STRUCTURES
 **
 *****************************************************************************/

/*
       ;*** FUNCTION AudioTest ***
       ;*** var
       ;AudioTest                       off=0000 key=93154 sz=02 ty=std ival=0
       ;rc                              off=0002 key=73829 sz=02 ty=std ival=0
       ;*** var_input
       ;id                              off=0004 key=93162 sz=02 ty=std ival=1
       ;playback                        off=0006 key=93168 sz=01 ty=std ival=1
       ;capture                         off=0007 key=93176 sz=01 ty=std ival=1
       ;*** var_input (access)
 */
typedef struct {
   /*** var */
   int16       retval;
   int16       error;
   /*** var_input */
   int16       id;
   int8        playback;
   int8        capture;

} ALIGN_ATTR tdef_audiotest;



/******************************************************************************
 **
 ** GLOBAL VARIABLES
 **
 *****************************************************************************/

static int8* playb_data=NULL;
static int   playb_size=0;


/******************************************************************************
 **
 ** LOCAL FUNCTIONS
 **
 *****************************************************************************/
#include <unistd.h>
/**
 * Loads the audio data from a file
 * @return 0 on success
 */
int load_playb_data(void){
   int rc;
   char path [200];
   FILE* fd;

   rc = vplFsMapFile(path, 200, "B:\\SND.WAV");
   if(rc){
      char msg[200];
      sprintf(msg, "Failed to create playback file path: %d", rc);
      vplDebug(msg);
      return -1;
   }
   fd = fopen(path, "r");
   if(fd == NULL){
      vplDebug("Failed to open playback file\n");
      return -1;
   }

   fseek(fd, 0, SEEK_END);
   size_t file_size = ftell(fd) - WAV_HEADER_LEN;

   rc = fseek(fd, WAV_HEADER_LEN, SEEK_SET);
   if(playb_data) free(playb_data);
   playb_data = malloc(file_size);
   if(!playb_data){
      vplDebug("Malloc failed\n");
      fclose(fd);
      return -2;
   }

   rc = fread(playb_data, 2, file_size/2, fd);
   if(rc>0){
      playb_size = rc*2;
   }else{
      playb_size = 0;
   }

   fclose(fd);
   return 0;
}

/**
 * Capture callback.
 * Copies the audio data from a file to the buffer.
 *
 * @param id            The ID of the registered stream.
 * @param buffer        Buffer to write the audio data to.
 * @param buffer_size   The size of the buffer
 * @param size          The actual written size
 * @return              0 on success
 */
static int test_stream_capt( __attribute__((unused)) int16 id, char* buffer, uint32 buffer_size, uint32* size){
   static int pos = 0;/// Keeps track of the current read position in the audio data.
   uint32 len;
   *size = 0;

   if(playb_data== NULL){
      // First run, load the data.
      load_playb_data();
   }
   if(!playb_data) return -1;

   if(pos>=playb_size) pos = 0;// End of data reached, start over from the start.

   len = playb_size-pos;
   if(len > buffer_size) len = buffer_size;

   memcpy(buffer, playb_data+pos, len);
   pos+=len;

   *size = len;

   return 0;
}

/**
 * Playback callback
 * Writes the received data to a file.
 *
 * @param id            The ID of the registered stream
 * @param data          The incoming audio data
 * @param size          The size of the incoming data.
 * @return              0 on success
 */
static int test_stream_playb(int16 id, char* data, uint32 size){
   char path[100];
   int rc;
   FILE* fd;

   rc = vplFsMapFile(path, 200, "B:\\REC_%d.WAV", id);
   if(rc){
      char msg[200];
      sprintf(msg, "Failed to create capture file path: %d", rc);
      vplDebug(msg);
      return -1;
   }

   fd = fopen(path, "a");

   if(fd == NULL){
      vplDebug("failed to open capture file");
      return -1;
   }

   fseek (fd, 0, SEEK_END);
   if(ftell(fd) == 0){
      // Empty file, write header
      uint32 tmp32;
      uint16 tmp16;

      char hdr[44];
      hdr[0] = 'R';
      hdr[1] = 'I';
      hdr[2] = 'F';
      hdr[3] = 'F';

      tmp32 = 0xffffffff;
      memcpy(&hdr[4], &tmp32, 4);

      hdr[8] = 'W';
      hdr[9] = 'A';
      hdr[10] = 'V';
      hdr[11] = 'E';

      hdr[12] = 'f';
      hdr[13] = 'm';
      hdr[14] = 't';
      hdr[15] = ' ';

      tmp32 = 16;
      memcpy(&hdr[16], &tmp32, 4);

      hdr[20] = 1;
      hdr[21] = 0;

      hdr[22] = 2;
      hdr[23] = 0;

      tmp32 = 16000;
      memcpy(&hdr[24], &tmp32, 4);

      tmp32 = 16000*2*16/8;
      memcpy(&hdr[28], &tmp32, 4);

      tmp16 = 2*16/8;
      memcpy(&hdr[32], &tmp16, 2);

      tmp16 = 16;
      memcpy(&hdr[34], &tmp16, 2);

      hdr[36] = 'd';
      hdr[37] = 'a';
      hdr[38] = 't';
      hdr[39] = 'a';

      tmp32 = 0xffffffff;
      memcpy(&hdr[40], &tmp32, 4);

      fwrite(hdr, 44, 1, fd);
   }

   fwrite(data, size, 1, fd);
   fclose(fd);

   return 0;
}



static int32 MODDECL audioRegisterStream( __attribute__((unused)) HANDLE* pCPU, __attribute__((unused)) void* pbase)
{
   // Var
   int rc;
   tdef_audiotest* test = (tdef_audiotest*)pbase;

   rc = vplAudioStreamRegister(test->id, test->capture?test_stream_capt:NULL, test->playback?test_stream_playb:NULL);

   // Completed
   return rc;
}


/******************************************************************************
 **
 ** FUNCTION TABLE
 **
 *****************************************************************************/
static vplFunctionEntry    ftable[] = { { "audioRegisterStream", audioRegisterStream }, { NULL, NULL } };


/******************************************************************************
 **
 ** EXPORTED FUNCTIONS
 **
 *****************************************************************************/

/////////////////////////////////////////////////////////////////////////////
// This function is called by the firmware when 'extModuleLoad' is called.
// Any initialization the module might require is performed here, including
// installing functions and function blocks.
//
// Input:
//   None.
//
// Output:
//   None.
//
// Returns:
//   0                  - Success
//   1                  - Failed
//
/////////////////////////////////////////////////////////////////////////////
int MODDECL moduleInit(void)
{
   // Var
   uint16 rc;

   // Initialize module
   rc = vplInstallFunctions("audioTest", ftable);
   if (rc)
   {
      // Install failed
      return 1;
   }

   // Completed
   return 0;
}

/////////////////////////////////////////////////////////////////////////////
// This function is called by the firmware to notify the module about an event.
//
// Input:
//   event              - The event ID.
//   parm               - The event parameter.
//
// Output:
//   None.
//
// Returns:
//   None.
//
/////////////////////////////////////////////////////////////////////////////
void MODDECL moduleNotify(int event, int parm)
{
   switch (event)
   {
      case EVENT_HALT:
         break;

      case EVENT_RESET:
         break;

      case EVENT_SHUTDOWN:
         break;

      case EVENT_POWERFAIL:
         if (parm) {
            // Running on battery
         } else {
            // Running on external power
         }
         break;

      case EVENT_POWERSAVE:
         if (parm) {
            // Suspend system
         } else {
            // Resume system
         }
         break;
   }
}


/******************************************************************************
 **
 **    EOF
 **
 *****************************************************************************/
