/** @file mod_mbus.c
 *  @brief Wireless M-Bus communication example module
 *
 *  This shows the framework necessary for communicating with the
 *  Radiocraft Wireless M-Bus RC1180-MBUS3 module.
 *
 *
 */

#include <module.h>

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <pthread.h>

/*****************************************************************************
 **
 **    DEFINITIONS AND MACROS
 **
 *****************************************************************************/
/// Maximum size of a MBUS message not including header information
#define MBUS_DATA_LENGTH	0xF6

/******************************************************************************
 **
 **    TYPEDEFS AND STRUCTURES
 **
 *****************************************************************************/

/**
 * @brief VPL data structures
 * @details Theses must match the output in the .lst file
 * @defgroup data_structures VPL data structures
 */
/**@{*/

/** @brief FUNCTION mbusOpen
 * <br/>
 * Defines the data structure of the VPL function mbusOpen.
 * This function opens the connection to the M-Bus module.
 */
struct ALIGN_ATTR mbus_open
{
   //*** var
   int16 retval; ///< return value from call, set by VPL
   int16 error; ///< error value from call, set by VPL

   //*** var_input
   int8 mode; ///< The M-Bus mode to use.
              ///< @li 0 = S2
              ///< @li 1 = T1 (default)
              ///< @li 2 = T2
              ///< @li 3 = S1
              ///< @li 4 = R
              ///< @li 10 = T1+C
              ///< @li 11 = T2+C
   int8 rssi; ///< Set to true to include the signal strength (RSSI) for each valid packet.
              ///< It can be read from mbusReceive() and is available in the mbusFrame struct block (mbus_frame).
//*** var_input (access)
};
/// Typedef for mbus_open
typedef struct mbus_open tdef_mbus_open;

/** @brief STRUCT_BLOCK mbusFrame
 * <br/>
 * Defines the data structure of the VPL STRUCT_BLOCK mbusFrame.
 * This struct block is used by mbusReceive() for storing the received frame.
 */
struct ALIGN_ATTR mbus_frame
{
   //*** var_input
   int8 control; ///< Control field
   int16 manufacture; ///< Manufacture ID. Part of the device address.
   int32 id; ///< Identification number/serial number. Part of the device address.
   int8 version; ///< Version number. Part of the device address.
   int8 type; ///< Device Type. Part of the device address.
   int16 rssi; ///< The signal strength (RSSI), if enabled in mbusOpen().
               ///< To get the signal strength in dB, P = -RSSI/2 [dBm]
   int16 length; ///< The length of the additional data in the frame.
   int8 data[MBUS_DATA_LENGTH]; ///< The additional data received from the M-Bus module. Only the first [#length] bytes are valid.
};

/// Typedef for mbus_frame
typedef struct mbus_frame tdef_mbus_frame;

/** @brief FUNCTION mbusSend
 * <br/>
 * Defines the data structure of the VPL function mbusSend.
 */
struct ALIGN_ATTR mbus_send
{
   //*** var
   int16 retval; ///< return value from call, set by VPL
   int16 error; ///< error value from call, set by VPL

   //*** var_input
   int16 control; ///< Value of the control field to use when sending.
   int16 length; ///< Length of the data to send
   int8* data; ///< Data to send

//*** var_input (access)
};

/// Typedef for mbus_send
typedef struct mbus_send tdef_mbus_send;

/** @brief FUNCTION mbusReceive
 * <br/>
 * Defines the data structure of the VPL function mbusReceive.
 * This function receives data from the M-Bus module.
 *
 */
struct ALIGN_ATTR mbus_receive
{
   //*** var
   int16 retval; ///< return value from call, set by VPL
   int16 error; ///< error value from call, set by VPL

   //*** var_input

   //*** var_input (access)
   tdef_mbus_frame* frame; ///< The provided struct is filled with the values of a received M-Bus frame.
};

/// Typedef for mbus_receive
typedef struct mbus_receive tdef_mbus_receive;

/** @brief FUNCTION mbusRegisterSlave
 * <br/>
 * Defines the data structure of the VPL function mbusRegisterSlave.
 * It registers/installs a slave device on the M-Bus module.
 *
 */
struct ALIGN_ATTR mbus_reg_slave
{
   //*** var
   int16 retval; ///< return value from call, set by VPL
   int16 error; ///< error value from call, set by VPL

   //*** var_input
   int16 idx; ///< The index of the slave to register. 1-64.
   int16 manufacturer; ///< Manufacturer ID. Part of the address of the slave.
   int32 id; ///< Identification number/serial number. Part of the address of the slave.
   int8 version; ///< Version number. Part of the address of the slave.
   int8 type; ///< Device Type. Part of the address of the slave.

   uint8* key; ///< Encryption key, 16 bytes. Can be set to NULL when encryption is not used.

//*** var_input (access)
};

/// Typedef for mbus_reg_slave
typedef struct mbus_reg_slave tdef_mbus_reg_slave;

/** @brief FUNCTION mbusSetFilter
 * <br/>
 * Defines the data structure of the VPL function mbusSetFilter.
 * Enables/disables filtering of the received packets, to only receive packets from registered slave devices.
 */
struct ALIGN_ATTR mbus_set_filter
{
   //*** var
   int16 retval; ///< return value from call, set by VPL
   int16 error; ///< error value from call, set by VPL

   //*** var_input
   int8 only_installed; ///< Set to true to only show packets from installed slaves.

//*** var_input (access)
};

/// Typedef for mbus_set_filter
typedef struct mbus_set_filter tdef_mbus_set_filter;

/**@}*/

/**
 * @brief The first block of data when a new MBUS message is received.
 * @details It contains the L-, C-, M- and A-fields and is used by mbusReceive() to store this information.
 */
struct PACKED_ATTR mbus_start_frame
{
   /// L-field : Length of full frame not including length or any start/stop bytes
   uint8 length;
   /// C-field : Control byte used to identify packet type
   uint8 control;
   /// M-Field : Manufacture ID, e.g. 0x0646 (ARF)
   uint16 manufacturer;
   /// A-Field[4]:ID/Serial number part of address field
   uint32 id;
   /// A-Field[5]:Device version part of address field
   uint8 version;
   /// A-Field[6]:Device type part of address field
   uint8 type;
};

/// Typedef for mbus_start_frame
typedef struct mbus_start_frame tdef_mbus_start_frame;

/******************************************************************************
 **
 ** LOCAL VARIABLES
 **
 *****************************************************************************/

/// serial port handle
static int fd = -1;

/// Flag to monitor if we are in config mode.
static int cfg_active = 0;
/// Flag to monitor if we have received the reply showing the we are in config mode.
static int cfg_activated = 0;
/// Flag to keep track of whether the RSSI is included in each packet.
static int rssi_included = 0;

/// thread synchronization lock
static pthread_mutex_t lock;

/******************************************************************************
 **
 ** LOCAL FUNCTIONS
 **
 *****************************************************************************/

/**
 * \defgroup helper_functions Helper C functions
 * \brief These functions are called from different C functions and contain common functionality.
 * @{
 */

/**
 * Send a debug message to the device output in the format "<prefix>:<formated string>"
 *
 * @param[in] prefix prefix to message
 * @param[in] format the format and arguments @see sprintf
 */
static void
mbusDebug(const char* prefix, const char *format, ...)
{
   va_list valist;
   char buffer[480];

   // build debug message;
   strcpy(buffer, prefix);
   strcat(buffer, ":");
   va_start(valist, format);
   vsprintf(&buffer[strlen(buffer)], format, valist);
   va_end(valist);

   vplDebug(buffer);
}

/**
 * @brief Control power to module
 *
 * Power the module on or off.
 *
 * @param[in] power If 0 then off else on.
 * @return -1 Failed to set power.
 * @return  0 Success.
 */
static int
mbusPower(uint8 power)
{
   if (vplSetIOSignal(RF_OFF, !power) != 0)
   {
      mbusDebug(__func__, "ERR (%d): Faield to set signal", __LINE__);
      return -1;
   }
   if (power)
   {
      // Disable CFG mode, as we are now powering the system
      vplSetIOSignal(RF_CFG, 0);
   }else{
      // Assert CFG pin to prevent it from keeping the module powered.
      vplSetIOSignal(RF_CFG, 1);
      // power turned off, mark as inactive
      cfg_activated = 0;
   }

   return 0;
}

/**
 * @brief Entering configuration mode
 *
 * Request entering configuration mode by sending '0x00', and wait for
 * confirmation. Any data received but not read and data written but not
 * transmitted will be flushed, to ensure synchronization.
 *
 * To exit configuration mode mbusConfigExit() must be called.
 *
 * @return -1 communication error
 * @return  0 success
 * @return  1 missing confirm
 */
static int
mbusConfigEnter(void)
{
   uint8 cmd = 0x00;
   uint8 reply = 0x00;
   int rc;

   if (cfg_active)
   {
      mbusDebug(__func__, "ERR (%d): Already in config mode", __LINE__);
      return -1;
   }

   // Deassert RF_CFG, to make it ready to assert it in the loop
   if (vplSetIOSignal(RF_CFG, 0) != 0)
   {
      mbusDebug(__func__, "ERR (%d): Failed to set signal", __LINE__);
      // We failed to set CFG signal, so we are fully depending on the 0x00 command.
   }

   for (int try = 0; try < 3; ++try)
   {
      // flush both data received but not read and data written but not transmitted.
      if (tcflush(fd, TCIOFLUSH))
      {
         mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
         return -1;
      }

      // flush send buffer
      if (tcdrain(fd))
      {
         mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
         return -1;
      }
      // set configuration mode request
      if (vplSetIOSignal(RF_CFG, 1) != 0)
      {
         mbusDebug(__func__, "ERR (%d): Failed to set signal", __LINE__);
      }

      if (!cfg_activated)
      {
         // send enter configuration mode request the first time, in case we already has entered it before, as this will resend the '>'.
         rc = write(fd, &cmd, 1);
         if (rc == -1)
         {
            mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
            return -1;
         }
         if (rc != 1)
         {
            mbusDebug(__func__, "ERR (%d): failed to send exit command",
                  __LINE__);
            return -1;
         }

      }
      usleep(50);

      // look for OK '>'
      reply = 0x00;
      rc = 0;
      int tries = 0;
      while (rc == 0 && tries < 3)
      {
         rc = read(fd, &reply, 1);
         if (rc == -1)
         {
            mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
            return -1;
         }
         tries++;
         usleep(10000);
      }

      // deassert config mode signal
      if (vplSetIOSignal(RF_CFG, 0) != 0)
      {
         mbusDebug(__func__, "ERR (%d): Failed to set signal", __LINE__);
      }

      if ((rc == 1) && reply == '>')
      {
         cfg_activated = 1;
         break;
      }
   }

   cfg_active = reply == '>';
   return (reply == '>') ? 0 : 1;
}

/**
 * @brief Exit configuration mode
 *
 * Request exit configuration mode by sending 'X', notice that the actual exit
 * can't be confirmed, and success is always assumed.
 *
 * To enter configuration mode, mbusConfigEnter() must be called.
 *
 * @return -1 communication error
 * @return  0 success
 */
static int
mbusConfigExit(void)
{
   uint8 cmd = 'X';
   if (!cfg_active)
   {
      mbusDebug(__func__, "ERR (%d): config not active", __LINE__);
      return -1;
   }

   // send exit request
   int rc = write(fd, &cmd, 1);
   if (rc == -1)
   {
      mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
      return -1;
   }
   if (rc != 1)
   {
      mbusDebug(__func__, "ERR (%d): failed to send exit command", __LINE__);
      return -1;

   }

   // flush send buffer
   if (tcdrain(fd))
   {
      mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
      return -1;
   }
   // flush recv buffer
   if (tcflush(fd, TCIOFLUSH))
   {
      mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
      return -1;
   }

   cfg_active = 0;
   return 0;
}

/**
 * @brief Send a single configuration command with additional data and receive the reply
 *
 * Will send the supplied configuration command/argument and the additional data, and return any
 * expected reply.<br>
 *
 * The module must first have been set in configuration mode using mbusConfigEnter().
 *
 * @param[in] cmd The command or value to send.
 * @param[in] data Additional data to send.
 * @param[in] d_size The length of the additional data to send
 * @param[in] reply Reference to where reply is stored.
 * @param[in] r_size Size of expected reply, '0' of no reply is expected.
 *
 * @return -1 communication error
 * @return  0 success
 * @return  1 missing confirm
 */
static int
mbusConfigTransact(uint8 cmd, uint8* data, int d_size, uint8* reply, int r_size)
{
   // write command
   int rc;
   rc = write(fd, &cmd, 1);
   if (rc == -1)
   {
      mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
      return -1;
   }
   if (rc != 1)
   {
      mbusDebug(__func__, "ERR (%d): failed to send 0x%02X", __LINE__, cmd);
      return -1;
   }

   if (data && d_size > 0)
   {
      rc = write(fd, data, d_size);
      if (rc == -1)
      {
         mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
         return -1;
      }
      if (rc != d_size)
      {
         mbusDebug(__func__, "ERR (%d): failed to send 0x%02X", __LINE__, cmd);
         return -1;
      }
   }

   // flush send buffer
   if (tcdrain(fd))
   {
      mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
      return -1;
   }

   // get reply
   if (r_size > 0)
   {
      int remaining = r_size;
      for (int try = 0; try < 5; try++)
      {
         int rc = read(fd, &reply[r_size - remaining], remaining);
         if (rc == -1)
         {
            mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
            return -1;
         }
         remaining -= rc;
         if (remaining == 0)
            break;
      }
      if (remaining)
      {
         mbusDebug(__func__,
               "ERR (%d): unexpected reply to 0x%02X (missing %d of %d)",
               __LINE__, (cmd & 0xFF), remaining, r_size);
         return -1;
      }
   }

   // wait for '>'
   uint8 cfm = 0;
   for (int try = 0; try < 5; try++)
   {
      rc = read(fd, &cfm, 1);
      if (rc == -1)
      {
         mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
         return -1;
      }
      if ((rc == 1) && cfm == '>')
         break;
   }
   // no OK reply
   if (cfm != '>')
   {
      mbusDebug(__func__, "ERR (%d): no reply to 0x%02X, 0x%x(%c)", __LINE__,
            cmd, cfm, cfm);
      return 1;
   }

   return 0;
}

/**
 * @brief Send a single command/argument
 *
 * Will send the supplied configuration command/argument, and return any
 * expected reply.<br>
 *
 * The module must first have been set in configuration mode using mbusConfigEnter().
 *
 * @param[in] cmd The command or value to send.
 * @param[in] reply Reference to where reply is stored.
 * @param[in] size Size of expected reply, '0' of no reply is expected.
 *
 * @return -1 communication error
 * @return  0 success
 * @return  1 missing confirm
 */
static int
mbusConfigCmd(uint8 cmd, uint8* reply, int size)
{
   return mbusConfigTransact(cmd, NULL, 0, reply, size);
}

/**
 * @brief Send configuration data
 *
 * Sends the supplied data and does not expect confirmation.<br/>
 *
 * The module must first have been set in configuration mode using mbusConfigEnter().
 *
 * @param[in] cmd The command or value to send.
 *
 * @return -1 communication error
 * @return  0 success
 */
static int
mbusConfigData(uint8 cmd)
{
   // write command
   int rc = write(fd, &cmd, 1);
   if (rc == -1)
   {
      mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
      return -1;
   }
   if (rc != 1)
   {
      mbusDebug(__func__, "ERR (%d): failed to send 0x%02X", __LINE__, cmd);
      return -1;
   }

   // flush send buffer
   if (tcdrain(fd))
   {
      mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
      return -1;
   }

   return 0;
}

/**
 * Reads the value of one or more registers in the M-Bus module and prints the value to the device output as hexadecimal values.
 * @param[in] reg The index of the first register to read
 * @param[in] size The number of sequential registers to read.
 * @param[in] name The name of the register to show in the output.
 * @return -1 communication error
 * @return  0 success
 */
static int mbus_read_reg(int8 reg, int size, const char* name)
{
   uint8 reply[20];
   char buf[60 + 1];
   if (size > 20)
      return -1;

   // Memory read + reg
   memset(reply, 0x00, size);
   for (int i = 0; i < size; ++i)
   {
      if (mbusConfigCmd('Y', NULL, 0))
         return -1;
      if (mbusConfigCmd(reg + i, &reply[i], 1))
         return -1;
      sprintf(&buf[i * 3], "%02X ", reply[i]);
   }

   mbusDebug(__func__, "%s = '%s'", name, buf);
   return 0;

}

/**
 * Reads the value of one or more registers in the M-Bus module and stores the value in the provided buffer.
 * @param[in] reg The index of the first register to read
 * @param[in] buffer The bufer to place the data in.
 * @param[in] size The number of sequential registers to read.
 * @return -1 communication error
 * @return  0 success
 */
static int mbus_read_regs(int8 reg, uint8* buffer, int size)
{

   // Memory read + reg
   memset(buffer, 0x00, size);
   for (int i = 0; i < size; ++i)
   {
      if (mbusConfigCmd('Y', NULL, 0))
         return -1;
      if (mbusConfigCmd(reg + i, &buffer[i], 1))
         return -1;
   }

   return 0;

}

/**
 * Change the value of a register in the M-Bus module.
 * It check to make sure that it does not already have the value, to avoid unneeded writing to the flash.
 * @param[in] reg The register to change
 * @param[in] val The new value
 * @return -1 communication error
 * @return  0 success, the register has the wanted value.
 */
static int mbus_config_mem_reg(uint8 reg, uint8 val)
{
   uint8 old_val;
   // Read before modify
   if (mbusConfigCmd('Y', NULL, 0))
      return -1;
   if (mbusConfigCmd(reg, &old_val, 1))
      return -1;

   if (old_val == val)
   {
      return 0;
   }

   // Memory modify
   if (mbusConfigCmd('M', NULL, 0))
      return -1;
   if (mbusConfigData(reg))
      return -1; // Addr
   if (mbusConfigData(val))
      return -1; // Val
   if (mbusConfigCmd(0xFF, NULL, 0))
      return -1; //0xff to end
   return 0;
}

/**
 * Configure a number of common settings.
 * @param[in] mode Set the M-Bus mode( see mbusOpen() for values). Set to -1 to not change it.
 * @param[in] rssi Set true to include RSSI in each packet. Set to -1 to not change it.
 * @param[in] only_installed Set true to only receive packets from installed/registered slaves. Set to -1 to not change it.
 * @return -1 communication error
 * @return  0 success
 */
static int mbus_config(int8 mode, int8 rssi, int8 only_installed)
{

   // request configuration mode
   if (mbusConfigEnter())
      return -1;

   if (mode >= 0)
   {
      // Memory modify + MBUS_MODE
      if (mbus_config_mem_reg(0x03, mode))
         return -1;
      mbus_read_reg(0x03, 1, "MBUS_MODE");
   }

   if (rssi >= 0)
   {
      // Memory modify + RSSI_MODE
      if (mbus_config_mem_reg(0x05, rssi ? 1 : 0))
         return -1; // Val

      mbus_read_reg(0x05, 1, "MBUS_RSSI");
      rssi_included = rssi;
   }

   if (only_installed >= 0)
   {
      // Memory modify + INSTALL_MODE
      if (mbus_config_mem_reg(0x3D, only_installed ? 0 : 2))
         return -1;
      mbus_read_reg(0x3d, 1, "MBUS_INSTALL_MODE");
   }

   // decrypt flag, always enabled
   if (mbus_config_mem_reg(0x3F, 1))
      return -1;

   // exit configuration mode and return to IDLE state
   if (mbusConfigExit())
      return -1;

   return 0;
}

/**
 * This function performs a factory reset of the RF module.
 * This includes erasing the serial number, so care should be taken when using it.
 *
 * @return -1 communication error
 * @return  0 success
 */
static int __attribute__( (unused))
mbusConfigFactoryReset(void)
{
   uint8 cmd = 0x00;
   uint8 reply = 0x00;
   int rc;

   if (cfg_active)
   {
      mbusDebug(__func__, "ERR (%d): Already in config mode", __LINE__);
      return -1;
   }

   if (vplSetIOSignal(RF_CFG, 0) != 0)
   {
      mbusDebug(__func__, "ERR (%d): Failed to set signal", __LINE__);
   }
   //sleep(1);
   for (int try = 0; try < 3; ++try)
   {
      // flush both data received but not read and data written but not transmitted.
      if (tcflush(fd, TCIOFLUSH))
      {
         mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
         return -1;
      }

      // flush send buffer
      if (tcdrain(fd))
      {
         mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
         return -1;
      }
      // set configuration mode request
      if (vplSetIOSignal(RF_CFG, 1) != 0)
      {
         mbusDebug(__func__, "ERR (%d): Failed to set signal", __LINE__);
      }

      if (!cfg_activated)
      {
         // send enter configuration mode request the first time, in case we already has entered it before, as this will resend the '>'.
         rc = write(fd, &cmd, 1);
         if (rc == -1)
         {
            mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
            return -1;
         }
         if (rc != 1)
         {
            mbusDebug(__func__, "ERR (%d): failed to send exit command",
                  __LINE__);
            return -1;
         }

      }
      usleep(50);

      // look for OK '>'
      reply = 0x00;
      rc = 0;
      int tries = 0;
      while (rc == 0 && tries < 3)
      {
         rc = read(fd, &reply, 1);
         if (rc == -1)
         {
            mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
            return -1;
         }
         tries++;
         usleep(10000);
      }

      if ((rc == 1) && reply == '>')
      {
         cfg_activated = 1;
         // In config mode: Send "reset to factory settings" request
         rc = mbusConfigTransact('@', (uint8*) "RC", 2, NULL, 0);

         // deassert config mode signal
         if (vplSetIOSignal(RF_CFG, 0) != 0)
         {
            mbusDebug(__func__, "ERR (%d): Failed to set signal", __LINE__);
         }
         break;
      }
      else
      {

         //sleep(1);
      }
      // deassert config mode signal
      if (vplSetIOSignal(RF_CFG, 0) != 0)
      {
         mbusDebug(__func__, "ERR (%d): Failed to set signal", __LINE__);
      }

   }

   cfg_active = reply == '>';

   return mbusConfigExit();

}
/**
 * @}
 */

/**
 * \defgroup impl_functions Implementation C functions
 * \brief These functions are called directly from the \ref modcall_functions.
 * @{
 */

/**
 * Closes the communication interface.
 * @return  0 success
 * @return -1 failed to close interface
 */
static int
mbusClose(void)
{
   if (fd >= 0)
   {
      // flush both data received but not read and data written but not transmitted.
      tcflush(fd, TCIOFLUSH);
      if (close(fd))
      {
         mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
         return -1;
      }
      mbusPower(0);
      fd = -1;
   }
   return 0;
}

/**
 * Opens the connection to the M-Bus module.
 * This must be called before the rest of the functions will work.
 *
 * @param[in] mode The M-Bus mode to use.
 * @li 0 = S2
 * @li 1 = T1 (default)
 * @li 2 = T2
 * @li 3 = S1
 * @li 4 = R
 * @li 10 = T1+C
 * @li 11 = T2+C
 * @param[in] rssi Set to true to include the signal strength (RSSI) for each valid packet.
 *              It can be read from mbusReceive() and is available in the mbusFrame struct block (mbus_frame).
 * @return  0 success
 * @return -1 failed to open interface
 */
static int
mbusOpen(int8 mode, int8 rssi)
{
   char port_path[200];
   struct termios tty;

   if (fd >= 0)
   {
      mbusDebug(__func__, "ERR: Communication interface already opened:");
      return -1;
   }

   if (vplFsMapFile(port_path, 200, "dev:ser3"))
   {
      mbusDebug(__func__, "ERR: Failed to open port:");
      return -1;
   }

   fd = open(port_path, O_RDWR | O_NOCTTY | O_SYNC);
   if (fd < 0)
   {
      mbusDebug(__func__, "ERR: Failed to open '%s': %s", port_path,
            strerror(errno));
      return -1;
   }

   memset(&tty, 0, sizeof tty);
   if (tcgetattr(fd, &tty) != 0)
   {
      mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
      return -1;
   }

   // default baud rate 19.2kBs
   cfsetospeed(&tty, B19200);
   cfsetispeed(&tty, B19200);

   tty.c_cflag |= (CLOCAL | CREAD); /* ignore modem controls */
   tty.c_cflag &= ~CSIZE;
   tty.c_cflag |= CS8; /* 8-bit characters */
   tty.c_cflag &= ~PARENB; /* no parity bit */
   tty.c_cflag &= ~CSTOPB; /* only need 1 stop bit */
   tty.c_cflag &= ~CRTSCTS; /* no hardware flow control */

   /* setup for non-canonical mode */
   tty.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL
         | IXON);
   tty.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
   tty.c_oflag &= ~OPOST;

   /* fetch bytes as they become available */
   tty.c_cc[VMIN] = (cc_t) 0; // non-blocking mode, only using timeout
   tty.c_cc[VTIME] = (cc_t) 10; // 1 seconds read timeout

   if (tcsetattr(fd, TCSANOW, &tty) != 0)
   {
      mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
   }

   if (mbusPower(1))
      return -1;

   // reset module by entering and exiting configuration mode
   int i;
   for (i = 0; i < 5; ++i)
   {
      if (mbusConfigEnter() == 0)
      {
         if (mbusConfigExit() == 0)
         {
            break;
         }
      }
   }
   if (i >= 5)
   {
      mbusDebug(__func__, "ERR (%d): failed to reset module", __LINE__);
   }

   if (mbus_config(mode, rssi, -1))
   {
      mbusPower(0);
      return -1;
   }

   return 0;
}

/**
 * @brief Send an M-BUS packet.
 *
 * This will send a M-BUS message, if c is not -1 then the control field will be
 * changed before sending. If no length is supplied then only the header frame
 * will be send.
 *
 * @param[in] c	New value for the Control field if not -1.
 * @param[in] data	Data to send, may be null if only header is to be send.
 * @param[in] size	Size of data to send, if 0 the only header will be send
 * @return -1 communication error
 * @return  0 success
 */
static int
mbusSend(int16 c, int8* data, int8 size)
{
   int send = 0;

   if (c >= 0)
   {
      // change control byte
      if (mbusConfigEnter())
         return -1;
      if (mbusConfigCmd('F', NULL, 0))
      {
         mbusConfigExit();
         return -1;
      }
      if (mbusConfigCmd(c, NULL, 0))
      {
         mbusConfigExit();
         return -1;
      }
      if (mbusConfigExit())
         return -1;
      // Delay 1.1ms from end of config to idle.
      usleep(1100);
   }

   // send length if given else request header only
   if (size == 0)
   {
      int cmd = 0xFE;
      // send only header information
      send = write(fd, &cmd, 1);
      if (send == -1)
      {
         mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
         return -1;
      }
      if (send != 1)
      {
         mbusDebug(__func__, "(%d): Failed to write send request", __LINE__);
         return -1;
      }
   }
   else
   {
      // send data length
      send = write(fd, &size, 1);
      if (send == -1)
      {
         mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
         return -1;
      }
      if (send != 1)
      {
         mbusDebug(__func__, "(%d): Failed to write data length", __LINE__);
         return -1;
      }
      send = write(fd, data, size);
      if (send == -1)
      {
         mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
         return -1;
      }
      if (send != size)
      {
         mbusDebug(__func__, "(%d): Failed to write data (%d of %d)", __LINE__,
               send, size);
         return -1;
      }
   }

   return 0;
}

/**
 * Receive a new frame from the M-Bus module
 * @param[in] start The frame to store the information from the header in.
 * @param[in] data Buffer to store the received data in.
 * @param[out] rssi If not NULL and if the RSSI is included in the packet,
 *             this is set to the RSSI of the packet.
 *
 * @return -1 communication error
 * @return  0 no data
 * @return  1 data received
 */
static int
mbusReceive(tdef_mbus_start_frame *start, int8 *data, int16* rssi)
{
   int rlen, data_size;
   int remaining = sizeof(*start);

   // get length
   rlen = read(fd, &start->length, 1);
   if (rlen < 0)
   {
      mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
      return -1;
   }
   if (rlen == 0)
   {
      // mbusDebug(__func__, "(%d): no length byte present", __LINE__);
      return 0; // no length present in receive buffer
   }
   if (start->length == 0)
   {
      mbusDebug(__func__, "(%d): zero length frame received", __LINE__);
      return 0; // no length present in receive buffer
   }

   remaining--;

   // get rest of start frame
   rlen = read(fd, (&start->length) + 1, remaining);
   if (rlen < 0)
   {
      mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
      return -1;
   }
   if (rlen != remaining)
   {
      mbusDebug(__func__, "(%d): failed to receive start frame (%d of %d)",
            __LINE__, rlen, remaining);
      tcflush(fd, TCIFLUSH); // flush in an attempt to recover
      return -1; // no length present in receive buffer
   }
   // get data
   data_size = start->length - sizeof(*start) + 1;

   if (data_size < 0)
   {
      mbusDebug(__func__, "(%d): invalid frame length (%d < %d)", __LINE__,
            start->length, sizeof(*start) + 1);
      tcflush(fd, TCIFLUSH); // flush in an attempt to recover
      return -1; // no length present in receive buffer
   }
   if (rssi_included)
   {
      data_size--;
   }
   if (data_size > 0)
   {
      int tries = 0;
      int pos = 0;
      memset(data, 0x00, MBUS_DATA_LENGTH);
      while (pos < data_size && tries < 3)
      {
         rlen = read(fd, &data[pos], data_size - pos);
         if (rlen < 0)
         {
            mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
            return -1;
         }
         pos += rlen;
         tries++;
         usleep(1000);
      }
      if (pos != data_size)
      {
         mbusDebug(__func__, "(%d): failed to receive data frame (%d of %d)",
               __LINE__, pos, data_size);
         tcflush(fd, TCIFLUSH); // flush in an attempt to recover
         return -1; // no length present in receive buffer
      }

   }
   if (rssi_included)
   {
      uint8 buf = 0;
      rlen = read(fd, &buf, 1);
      if (rlen < 0)
      {
         mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
         return -1;
      }
      if (rlen == 1)
      {
         if (rssi)
         {
            *rssi = buf;
         }
      }

   }
   start->length = (uint8) data_size;

   return 1;
}

/**
 * Registers/installs a slave device on the M-Bus module.
 * This is necessary when filtering or when using encryption.
 *
 * @param[in] idx The index of the slave to register. 1-64.
 * @param[in] manufacturer The ID of the manufacturer. Part of the device address.
 * @param[in] id The device identification number/serial number.  Part of the device address.
 * @param[in] version The device version. Part of the device address.
 * @param[in] type The device type. Part of the device address.
 * @param[in] key The 16 byte (128 bit) encryption key. Can be set to NULL if encryption is not used.
 * @return -2 Invalid idx
 * @return -1 communication error
 * @return  0 Slave was registered
 */
static int
mbus_register_slave(int idx, int16 manufacturer, int32 id, int8 version,
                    int8 type,
                    uint8* key)
{
   uint8 addr[8];

// curent slave index
   if (idx > 64 || idx <= 0)
   {
      return -2;
   }

   memcpy(&addr[0], &id, 4);
   memcpy(&addr[4], &manufacturer, 2);
   memcpy(&addr[6], &version, 1);
   memcpy(&addr[7], &type, 1);

   // request configuration mode
   if (mbusConfigEnter())
      return -1;

   // bind slave
   if (mbusConfigCmd('B', NULL, 0))
      return -1;
   if (mbusConfigTransact(idx, addr, 8, NULL, 0))
      return -1;

   if (key)
   {
      // set key for slave
      if (mbusConfigCmd('K', NULL, 0))
         return -1;
      if (mbusConfigTransact(idx, key, 16, NULL, 0))
         return -1; // Slave ID
   }

   // exit configuration mode and return to IDLE state
   if (mbusConfigExit())
      return -1;

   return 0;
}

/**
 * @brief Print some information about the RF module to the device output.
 *
 * @return -1 Communication error.
 * @return  0 Success.
 */
static int
mbus_info(void)
{
   uint8 reply[20];
   char serial[8 * 3];
   int i;

   // request configuration mode
   if (mbusConfigEnter())
      return -1;

   // Signal strength
   if (mbusConfigCmd('Q', reply, 1))
      return -1;
   mbusDebug(__func__, "Signal quality = %d", reply[0]);

   // Signal strength
   if (mbusConfigCmd('S', reply, 1))
      return -1;
   mbusDebug(__func__, "RSSI           = %d (%d dBm)", reply[0],
         ((-reply[0]) / 2));

   // Temperature monitoring
   if (mbusConfigCmd('U', reply, 1))
      return -1;
   mbusDebug(__func__, "Temperature    = %d C", (reply[0] - 128));

   // Battery Monitoring
   if (mbusConfigCmd('V', reply, 1))
      return -1;
   mbusDebug(__func__, "VCC            = %dmV", (reply[0] * 30));

   // Memory read + MBUS_MODE
   if (mbusConfigCmd('Y', NULL, 0))
      return -1;
   if (mbusConfigCmd(0x03, reply, 1))
      return -1;
   mbusDebug(__func__, "MBUS_MODE      = 0x%02X", reply[0]);

   // Memory read + DATA_INTERFACE
   if (mbus_read_regs(0x36, reply, 1))
      return -1;
   mbusDebug(__func__, "DATA_INTERFACE = 0x%02X", reply[0]);

   // Memory read + PART_NUMBER
   if (mbus_read_regs(0x61, reply, 12))
      return -1;
   reply[12] = 0;
   mbusDebug(__func__, "PART_NUMBER    = '%s'", reply);

   // Memory read + HW_REV_NO
   if (mbus_read_regs(0x6E, reply, 4))
      return -1;
   reply[4] = 0;
   mbusDebug(__func__, "HW_REV_NO      = '%s'", reply);

   // Memory read + FW_REV_NO
   if (mbus_read_regs(0x73, reply, 4))
      return -1;
   reply[4] = 0;
   mbusDebug(__func__, "FW_REV_NO      = '%s'", reply);

   // Memory read + SERIAL_NUMBER
   if (mbus_read_regs(0x78, reply, 8))
      return -1;
   for (i = 0; i < 8; i++)
   {
      sprintf(&serial[i * 3], "%02X ", reply[i]);
   }
   serial[8 * 3 - 1] = 0;
   mbusDebug(__func__, "SERIAL_NUMBER  = %s", serial);

   if (mbus_read_regs(0x3d, reply, 1))
      return -1;
   mbusDebug(__func__, "INSTALL_MODE   = 0x%02X", reply[0]);

   if (mbus_read_regs(0x05, reply, 1))
      return -1;
   mbusDebug(__func__, "RSSI_MODE      = 0x%02X", reply[0]);

   // exit configuration mode and return to IDLE state
   if (mbusConfigExit())
      return -1;

   return 0;
}

/**
 * @}
 */

/**
 * \defgroup modcall_functions Modcall C functions
 * \brief These functions can be called from VPL using the MODCALL keyword
 * @{
 */

/**
 * Open communication interface to the module.
 * @param[in] pCPU	The handle to the system.
 * @param[in] pBase	Pointer to structure with function parameters.
 */
static int32 MODDECL
mbusCallOpen(HANDLE* pCPU __attribute__( (unused)),
             void* pBase)
{
   tdef_mbus_open* pdata = (tdef_mbus_open*) pBase;
   if (pthread_mutex_lock(&lock))
   {
      mbusDebug(__func__, "ERR (%d): failed to get lock", __LINE__);
      return -1;
   }

   int rc = mbusOpen(pdata->mode, pdata->rssi);
   pthread_mutex_unlock(&lock);

   return rc;
}

/**
 * Close communication interface to the module.
 * @param[in] pCPU	The handle to the system.
 * @param[in] pBase	Pointer to structure with function parameters.
 */
static int32 MODDECL
mbusCallClose(HANDLE* pCPU __attribute__((unused)),
              void* pBase __attribute__((unused)))
{
   if (pthread_mutex_lock(&lock))
   {
      mbusDebug(__func__, "ERR (%d): failed to get lock", __LINE__);
      return -1;
   }

   int rc = mbusClose();
   pthread_mutex_unlock(&lock);

   return rc;
}

/**
 * @brief Check if any MBUS packets has been received
 * @param[in] pCPU	The handle to the system.
 * @param[in] pBase pointer to VPL functions data structure
 */
static int32 MODDECL
mbusCallReceive(HANDLE* pCPU __attribute__((unused)), void* pBase)
{
   int16 rssi = 0;
   tdef_mbus_start_frame head;
   tdef_mbus_receive *pdata = (tdef_mbus_receive*) pBase;

   if (pthread_mutex_lock(&lock))
   {
      mbusDebug(__func__, "ERR (%d): failed to get lock", __LINE__);
      return -1;
   }

   int rc = mbusReceive(&head, pdata->frame->data, &rssi);
   if (rc > 0)
   {
      pdata->frame->control = head.control;

      pdata->frame->control = (int8) head.control;
      write16b(&pdata->frame->manufacture, head.manufacturer);
      write32b(&pdata->frame->id, head.id);
      pdata->frame->version = (int8) head.version;
      pdata->frame->type = head.type;
      write16b(&pdata->frame->rssi, rssi);
      write16b(&pdata->frame->length, head.length);
   }

   pthread_mutex_unlock(&lock);

   return rc;
}

/**
 * @brief Send a MBUS message
 * @param[in] pCPU	The handle to the system.
 * @param[in] pBase pointer to VPL functions data structure
 * @return 0 Message sent
 * @return -1 Invalid parameter.
 * @return -2 Failed to send message
 */
static int32 MODDECL
mbusCallSend(HANDLE* pCPU __attribute__((unused)), void* pBase)
{
   tdef_mbus_send *pdata = (tdef_mbus_send*) pBase;

   if ((pdata->length > 0) && !pdata->data)
   {
      mbusDebug(__func__, "ERR (%d): missing data", __LINE__);
      return -1;
   }

   if ((pdata->length < 0) || (pdata->length > MBUS_DATA_LENGTH))
   {
      mbusDebug(__func__, "ERR (%d): illegal data length %d", __LINE__,
            pdata->length);
      return -1;
   }

   if ((pdata->control < -1) || (pdata->control > 0xFF))
   {
      mbusDebug(__func__, "ERR (%d): illegal control value %d", __LINE__,
            pdata->length);
      return -1;
   }

   if (pthread_mutex_lock(&lock))
   {
      mbusDebug(__func__, "ERR (%d): failed to get lock", __LINE__);
      return -1;
   }

   int rc = mbusSend(pdata->control, pdata->data, pdata->length);

   pthread_mutex_unlock(&lock);
   return rc ? -2 : 0;
}

/**
 * @brief Perform a series of tests to validate communication
 *
 * Perform a series of test by reading out configuration values to validate
 * communication to module, result is send to device output.
 *
 * Prints:
 * @li	Signal strength
 * @li	Signal quality
 * @li	Temperature monitoring
 * @li	Battery Monitoring
 * @li	MBUS_MODE
 * @li	DATA_INTERFACE
 * @li	PART_NUMBER
 * @li	HW_REV_NO
 * @li	FW_REV_NO
 * @li	SERIAL_NUMBER
 * @li   INSTALL_MODE
 * @li   RSSI_MODE
 *
 * Result
 @verbatim
 mbus_test:Signal quality = 0
 mbus_test:RSSI           = 205 (-102 dBm)
 mbus_test:Temperature    = 32 C
 mbus_test:VCC            = 3450mV
 mbus_test:MBUS_MODE      = 0x01
 mbus_test:DATA_INTERFACE = 0x00
 mbus_test:PART_NUMBER    = RC1180-MBUS3
 mbus_test:HW_REV_NO      = 2.00
 mbus_test:FW_REV_NO      = 3.15
 mbus_info:SERIAL_NUMBER  = 01 23 45 67 89 AB CD EF
 mbus_info:INSTALL_MODE   = 0x02
 mbus_info:RSSI_MODE      = 0x01

 @endverbatim
 * @param[in] pCPU	The handle to the system.
 * @param[in] pBase pointer to VPL functions data structure
 * @return 0 Message send
 * @return -1 Failed to send message
 */
static int32 MODDECL
mbusCallInfo(HANDLE* pCPU __attribute__((unused)),
             void* pBase __attribute__((unused)))
{
   if (pthread_mutex_lock(&lock))
   {
      mbusDebug(__func__, "ERR (%d): failed to get lock", __LINE__);
      return -1;
   }

   int rc = mbus_info();
   if (rc)
   {
      mbusDebug(__func__, "error resetting interface (%d)", rc);
      mbusClose();
      mbusOpen(-1, 0);
   }

   pthread_mutex_unlock(&lock);
   return 0;
}

/**
 * Set mode
 * @param[in] pCPU The handle to the system.
 * @param[in] pBase   Pointer to structure with function parameters.
 */
static int32 MODDECL
mbusCallSetFilter(HANDLE* pCPU __attribute__( (unused)),
                  void* pBase)
{
   tdef_mbus_set_filter* pdata = (tdef_mbus_set_filter*) pBase;
   if (pthread_mutex_lock(&lock))
   {
      mbusDebug(__func__, "ERR (%d): failed to get lock", __LINE__);
      return -1;
   }

   int rc = mbus_config(-1, -1, pdata->only_installed);
   pthread_mutex_unlock(&lock);

   return rc;
}

/**
 * Register a slave device. It is stored persistently.
 * @param[in] pCPU The handle to the system.
 * @param[in] pBase   Pointer to structure with function parameters.
 */
static int32 MODDECL
mbusCallRegisterSlave(HANDLE* pCPU __attribute__( (unused)),
                      void* pBase)
{
   tdef_mbus_reg_slave* pdata = (tdef_mbus_reg_slave*) pBase;
   if (pthread_mutex_lock(&lock))
   {
      mbusDebug(__func__, "ERR (%d): failed to get lock", __LINE__);
      return -1;
   }

   int rc = mbus_register_slave(pdata->idx, pdata->manufacturer, pdata->id,
         pdata->version, pdata->type, pdata->key);
   pthread_mutex_unlock(&lock);

   return rc;
}

/** @} */

/**
 * \defgroup exported_functions Exported functions
 * \brief These functions are required to allow the module to be loaded from VPL
 * @{
 */

/**
 * This function is called by the firmware when 'extModuleLoad' is called.
 * Any initialization the module might require is performed here, including
 * installing functions and function blocks.
 *
 * @return 0	Success
 * @return 1	Failed
 */
int
moduleInit(void)
{
   // Var
   uint16 rc;
   pthread_mutexattr_t lock_attr;
   vplFunctionEntry ftable[] =
         {
               { "mbusOpen", mbusCallOpen },
               { "mbusClose", mbusCallClose },
               { "mbusSend", mbusCallSend },
               { "mbusReceive", mbusCallReceive },
               { "mbusInfo", mbusCallInfo },
               { "mbusRegisterSlave", mbusCallRegisterSlave },
               { "mbusSetFilter", mbusCallSetFilter },
               {
                     NULL, NULL } };

   mbusDebug(__func__, "Initializing");

   // initialize API lock
   rc = pthread_mutexattr_init(&lock_attr);
   if (rc != 0)
   {
      mbusDebug(__func__, "ERR (%d): %d", __LINE__, rc);
      return 1;
   }
   rc = pthread_mutexattr_settype(&lock_attr, PTHREAD_MUTEX_RECURSIVE);
   if (rc != 0)
   {
      mbusDebug(__func__, "ERR (%d): %d", __LINE__, rc);
      return 1;
   }
   rc = pthread_mutex_init(&lock, &lock_attr);
   if (rc != 0)
   {
      mbusDebug(__func__, "ERR (%d): %d", __LINE__, rc);
      return 1;
   }

   // Initialize module
   rc = vplInstallFunctions("mod_mbus", ftable);
   if (rc)
   {
      // Install failed
      mbusDebug(__func__, "failed to install 'mod_mbus' functions err %d", rc);
      return 1;
   }

   if (vplSetIOSignal(RF_CFG, 0) != 0)
   {
      mbusDebug(__func__, "ERR (%d): %s", __LINE__, strerror(errno));
   }

   if (mbusPower(0))
      return -1;

   // Completed
   return 0;
}

/**
 * @brief This function is called by the firmware to notify the module about an event.
 *
 * In case of a halt, reset or shutdown event is revived then the interface
 * is closed by calling mbusClose().
 *
 * @param[in] event The event ID.
 * @param[in] parm The event parameter.
 **/

/**
 *
 */
void
moduleNotify(int event, int parm)
{
   switch (event)
   {
   case EVENT_HALT:
      case EVENT_RESET:
      case EVENT_SHUTDOWN:
      mbusClose();
      break;

   case EVENT_POWERFAIL:
      if (parm)
      {
         // Running on battery
      }
      else
      {
         // Running on external power
      }
      break;

   case EVENT_POWERSAVE:
      if (parm)
      {
         // Suspend system
      }
      else
      {
         // Resume system
      }
      break;

   default:
      mbusDebug(__func__, "Unknown event %d", event);
      break;
   }
}
/**@}*/

/******************************************************************************
 **
 **    EOF
 **
 *****************************************************************************/
