//-----------------------------------------------------------------------------
// Wireless M-Bus communication
//-----------------------------------------------------------------------------
#IFDEF NOT MBUS_INC THEN
#DEFINE MBUS_INC

#DEFINE MBUS_MODULE_MEDIA  1
#DEFINE MBUS_MODULE_PATH   "B:\SYSTEM\EXT\MOD_MBUS.RMX"

#DEFINE MBUS_DATA_LENGTH   246

#DEFINE MBUS_MODCALL_ERR   -999

#DEFINE MBUS_MODE_S2       0
#DEFINE MBUS_MODE_T1       1
#DEFINE MBUS_MODE_T2       2
#DEFINE MBUS_MODE_S1       3
#DEFINE MBUS_MODE_R        4
#DEFINE MBUS_MODE_T1_C     10
#DEFINE MBUS_MODE_T2_C     11


//----------------------------------------------------------------------------
// mbusFrame
//
// Description:
//    This struct block contains a frame received from the M-Bus module.
//
// Output:
//    control      - The control field.
//    manufacturer - The ID of the manufacturer. Part of the device address.
//    id           - The device id/serial number.  Part of the device address.
//    version      - The device version. Part of the device address.
//    type         - The device type. Part of the device address.
//    RSSI         - The signal strength for the packet. If it is not enabled in mbusOpen, it is set to 0.
//    length       - The length of the additional data in the frame.
//    data         - The additional data that was received. 
//                   Only the first <length> bytes are valid.
//
//----------------------------------------------------------------------------
STRUCT_BLOCK ALIGN mbusFrame;
   control      : SINT;           // Control field
   
   // Address
   manufacturer : INT;           // Manufacture ID
   id           : DINT;          // Device Identification number
   version      : SINT;          // Device Version number
   type         : SINT;          // Device Type

   // APP_DATA
   RSSI         : INT;            // signal strength in dB = - RSSI/2
   length       : INT;            // Length of additional data
   data         : ARRAY[1 .. MBUS_DATA_LENGTH] OF SINT;  
END_STRUCT_BLOCK;

//----------------------------------------------------------------------------
// mbusInit()
//
// Description:
//    Initializes the M-Bus extension module.
//
// Input:
//    showTags   - Print the tags from the module to the device output.
//    path       - The path to the module to initialize. Optional, as the default value is already correct.
//    media      - The media the module is located on. Optional, as the default value is already correct.
//
// Returns:
//        0   : OK
//       -1   : Failed to open media
//       -2   : Failed show tags
//       -3   : Failed to load module
//----------------------------------------------------------------------------
FUNCTION ALIGN mbusInit : INT;
VAR_INPUT
   showTags : BOOL := FALSE;
   path     : STRING := MBUS_MODULE_PATH;
   media    : SINT := MBUS_MODULE_MEDIA;
END_VAR;
VAR
   rc       : INT;
   index    : INT;   
   text     : STRING;
   tag      : STRING;
END_VAR;

// Initialize media where module is located
rc := fsMediaOpen(media := media);
IF NOT (rc = 0) THEN
   DebugFmt(message := "Failed to open media \1 (err \2)", v1 := media, v2 := rc);
   mbusInit := -1;
   RETURN;
END_IF;

IF showTags THEN
   DebugMsg(message := "Extension:     " + path);
   FOR index := 1 TO 10 DO
      // Read the name of a tag
      rc:=extTagEnumerate(path:=path, index:=index, tag:=tag);
      IF (rc = 1) THEN
         // Read the text of a tag
         rc := extTagRead(path:=path,tag:=tag,text:=text);
         IF (rc = 1) THEN
            DebugFmt(message := "[\1] : '" + strLeft(str := tag + "'            ", length := 13) + "= '" + text + "'", v1 := index);
         ELSE
            DebugFmt(message := "[\1] : Failed to read '" + tag + "' (err \2)", v1 := index, v2 := rc);
         END_IF;
      ELSIF (rc = -3) THEN
         // No more tags
         EXIT;
      ELSE
         DebugFmt(message := "[\1] : Failed to read tag name (err \2)", v1 := index, v2 := rc);
         mbusInit := -2;
         RETURN;
      END_IF;   
   END_FOR;
END_IF;

// Load an extension module
rc := extModuleLoad(path:=path);
IF NOT (rc = 1) THEN
   DebugFmt(message := "Failed to load module " + path + "(err \1)", v1 := rc);
   mbusInit := -3;
   RETURN;
END_IF;

DebugMsg(message := "Module '" + path + "' loaded");
mbusInit := 0;
END_FUNCTION;

//----------------------------------------------------------------------------
// mbusOpen()
//
// Description:
//    Opens the connection to the M-Bus module.
//
// Input:
//    mode       - The M-Bus mode to use. See the MBUS_MODE_* defines above.
//    RSSI       - Include the signal strength( RSSI) for each valid packet.
//                 It can be read from mbusReceive()
//
// Returns:
//        0: Success
//       -1: Failed to open the M-Bus interface.
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION ALIGN mbusOpen : INT;
VAR_INPUT
   mode: SINT := MBUS_MODE_T1;
   RSSI: BOOL := OFF;
END_VAR;
VAR
   error    : INT;
END_VAR;
   mbusOpen := INT(MODCALL("mod_mbus", "mbusOpen", error));
   IF NOT (error = 0) THEN
      DebugFmt(message := "MBUS: failed to call external module 'mbusOpen' (err \1)", v1 := error);
      mbusOpen := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;

//----------------------------------------------------------------------------
// mbusClose()
//
// Description:
//    Closes the connection to the M-Bus module.
//
// Input:
//    None.
//
// Returns:
//        0: Success
//       -1: Failed to close the M-Bus interface.
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION ALIGN mbusClose : INT;
VAR
   error    : INT;
END_VAR;
   mbusClose := INT(MODCALL("mod_mbus", "mbusClose", error));
   IF NOT (error = 0) THEN
      DebugFmt(message := "MBUS: failed to call external module 'mbusClose' (err \1)", v1 := error);
      mbusClose := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;

//----------------------------------------------------------------------------
// mbusSend()
//
// Description:
//    Sends data to the M-Bus module.
//
// Input:
//    control    - The M-Bus mode to use. See the MBUS_MODE_* defines above. 0-255.
//    length     - The length of the data to send. Max 246 bytes.
//    data       - The data to send.
//
// Returns:
//        0: Success
//       -1: Invalid parameter.
//       -2: Failed to send data.
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION ALIGN mbusSend : INT;
VAR_INPUT
   control     : INT := -1;   // (-1, 0 .. 255) :change control byte before send
   length      : INT := 0;    // (0 .. MBUS_DATA_LENGTH) : length of data to send
   data        : PTR;         // data to send
END_VAR;
VAR
   error       : INT;
END_VAR;
   mbusSend := INT(MODCALL("mod_mbus", "mbusSend", error));
   IF NOT (error = 0) THEN
      DebugFmt(message := "MBUS: failed to call external module 'mbusSend' (err \1)", v1 := error);
      mbusSend := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;

//----------------------------------------------------------------------------
// mbusReceive()
//
// Description:
//    Receives data from the M-Bus module.
//
// Input:
//    None.
//
// Output:
//    frame      - The provided struct is filled with the values of a received M-Bus frame. 
//
// Returns:
//        1: Data was received.
//        0: No data
//       -1: Communicaton error
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION ALIGN mbusReceive;
VAR_INPUT
   frame : ACCESS mbusFrame;   
END_VAR;
VAR
   error       : INT;
END_VAR;
   mbusReceive := INT(MODCALL("mod_mbus", "mbusReceive", error));
   IF NOT (error = 0) THEN
      DebugFmt(message := "MBUS: failed to call external module 'mbusReceive' (err \1)", v1 := error);
      mbusReceive := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;

//----------------------------------------------------------------------------
// mbusInfo()
//
// Description:
//    Sends some information about the M-Bus module to the device output.
// Example output:
//
//    mbus_test:Signal quality = 0
//    mbus_test:RSSI           = 205 (-102 dBm)
//    mbus_test:Temperature    = 32 C
//    mbus_test:VCC            = 3450mV
//    mbus_test:MBUS_MODE      = 0x01
//    mbus_test:DATA_INTERFACE = 0x00
//    mbus_test:PART_NUMBER    = RC1180-MBUS3
//    mbus_test:HW_REV_NO      = 2.00
//    mbus_test:FW_REV_NO      = 3.15
//    mbus_info:SERIAL_NUMBER  = 01 23 45 67 89 AB CD EF
//    mbus_info:INSTALL_MODE   = 0x02
//    mbus_info:RSSI_MODE      = 0x01
//
// Input:
//    None.
//
// Returns:
//        0: Success
//       -1: Communicaton error.
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION mbusInfo : INT;
VAR
   error    : INT;
END_VAR;
   mbusInfo := INT(MODCALL("mod_mbus", "mbusInfo", error));
   IF NOT (error = 0) THEN
      DebugFmt(message := "MBUS: failed to call external module 'mbusInfo' (err \1)", v1 := error);
      mbusInfo := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;


//----------------------------------------------------------------------------
// mbusRegisterSlave()
//
// Description:
//    Registers/installs a slave device on the M-Bus module.
//    This is necessary when filtering or when using encryption.
//
// Input:
//    idx          - The index of the slave to register. 1-64.
//    manufacturer - The ID of the manufacturer. Part of the device address.
//    id           - The device identification number/serial number.  Part of the device address.
//    version      - The device version. Part of the device address.
//    type         - The device type. Part of the device address.
//    key          - The 16 byte (128 bit) encryption key. Can be left empty if encryption is not used.
//
// Returns:
//        0: Success
//       -1: Communicaton error.
//       -2: Invalid idx.
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION mbusRegisterSlave : INT;
VAR_INPUT
   idx           : INT := 1;
   manufacturer  : MANDATORY INT;
   id            : MANDATORY DINT;
   version       : MANDATORY SINT;
   type          : MANDATORY SINT;
   key           : PTR;
END_VAR;
VAR
   error    : INT;
END_VAR;
   mbusRegisterSlave := INT(MODCALL("mod_mbus", "mbusRegisterSlave", error));
   IF NOT (error = 0) THEN
      DebugFmt(message := "MBUS: failed to call external module 'mbusRegisterSlave' (err \1)", v1 := error);
      mbusRegisterSlave := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;

//----------------------------------------------------------------------------
// mbusSetFilter()
//
// Description:
//    Enables/disables filtering of the received packets, to only receive packets from registered slave devices.
//
// Input:
//    installed    - Set to true to only show packets from installed slaves.
//
// Returns:
//        0: Success
//       -1: Communicaton error.
//       -2: Invalid idx.
//     -999 (MBUS_MODCALL_ERR): Failed to find function in extension module.
//----------------------------------------------------------------------------
FUNCTION mbusSetFilter : INT;
VAR_INPUT
   installed           : BOOL := TRUE;
END_VAR;
VAR
   error    : INT;
END_VAR;
   mbusSetFilter := INT(MODCALL("mod_mbus", "mbusSetFilter", error));
   IF NOT (error = 0) THEN
      DebugFmt(message := "MBUS: failed to call external module 'mbusSetFilter' (err \1)", v1 := error);
      mbusSetFilter := MBUS_MODCALL_ERR;
   END_IF;
END_FUNCTION;

#END_IF