/***************************** C SOURCE FILE **********************************
 **
 **  Project:    VPL extension module
 **  Filename:   ntp_mod.c
 **
 ******************************************************************************
 **
 **  This exports a function for retrieving the current time from an NTP server.
 **
 *****************************************************************************/

/******************************************************************************
 **
 ** MODULES USED
 **
 *****************************************************************************/
#include <module.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include "ntp.h"

/******************************************************************************
 **
 **    TYPEDEFS AND STRUCTURES
 **
 *****************************************************************************/

typedef struct {
   /*** var */
   int32       retval;
   int16       error;
   int16       host;
   int16       tz;

} ALIGN_ATTR tdef_get_time;


/******************************************************************************
 **
 ** LOCAL FUNCTIONS
 **
 *****************************************************************************/

static int32 MODDECL get_time(HANDLE* pCPU, void* pbase)
{
   // Var
   time_t ts;
   struct tm *t;
   tdef_get_time* gt = (tdef_get_time*)pbase;

   const char* host = vplStringGet(pCPU, gt->host);
   const char* tz = vplStringGet(pCPU, gt->tz);

   ts = ntp_get_time(host);
   if(ts>0){
      if(tz && strlen(tz) > 0){
      	// Set TZ variable to calculate local time
         if(setenv("TZ", tz, 1)){
            return -2;
         }
         tzset();
      }
      t = localtime(&ts);
      
      if(tz && strlen(tz) > 0){
      	// Unset TZ again to avoid affecting the rest of the system.
         unsetenv("TZ");
         tzset();
      }
      // Return the calculated time
      return vplClockTMToLinsec(t);
   }
   return -1;
}



/******************************************************************************
 **
 ** FUNCTION TABLE
 **
 *****************************************************************************/
static vplFunctionEntry    ftable[] = { { "time", get_time }, { NULL, NULL } };


/******************************************************************************
 **
 ** EXPORTED FUNCTIONS
 **
 *****************************************************************************/

/////////////////////////////////////////////////////////////////////////////
// This function is called by the firmware when 'extModuleLoad' is called.
// Any initialization the module might require is performed here, including
// installing functions and function blocks.
//
// Input:
//   None.
//
// Output:
//   None.
//
// Returns:
//   0                  - Success
//   1                  - Failed
//
/////////////////////////////////////////////////////////////////////////////
int MODDECL moduleInit(void)
{
   // Var
   uint16 rc;

   // Initialize module
   rc = vplInstallFunctions("ntp", ftable);
   if (rc)
   {
      // Install failed
      return 1;
   }

   // Completed
   return 0;
}

/////////////////////////////////////////////////////////////////////////////
// This function is called by the firmware to notify the module about an event.
//
// Input:
//   event              - The event ID.
//   parm               - The event parameter.
//
// Output:
//   None.
//
// Returns:
//   None.
//
/////////////////////////////////////////////////////////////////////////////
void MODDECL moduleNotify(int event, int parm)
{
   switch (event)
   {
      case EVENT_HALT:
         break;

      case EVENT_RESET:
         break;

      case EVENT_SHUTDOWN:
         break;

      case EVENT_POWERFAIL:
         if (parm) {
            // Running on battery
         } else {
            // Running on external power
         }
         break;

      case EVENT_POWERSAVE:
         if (parm) {
            // Suspend system
         } else {
            // Resume system
         }
         break;
   }
}


/******************************************************************************
 **
 **    EOF
 **
 *****************************************************************************/
